import pandas as pd
import numpy as np


class Prediction:

    param_model_map = {"arima": "ARIMA", "linear": "LINEAR", "svr": "SVR"}
    # either all to predict using all the models else one of arima, svr or linear
    prediction_type = None
    # input data
    input_df = None
    # the column name to be used as the date column
    date_column = None
    # The value to be predicted
    label = None
    # the number of data points in the future to be predicted, max 1000
    data_points = 20
    model_params = dict(
        arima=dict(
            start_p=1,
            start_q=1,
            test="adf",  # use adftest to find optimal 'd'
            max_p=3,
            max_q=3,  # maximum p and q
            m=1,  # frequency of series
            d=None,  # let model determine 'd'
            seasonal=False,  # No Seasonality / True for seasoanal i.e. SARIMA
            start_P=0,
            D=0,
            trace=True,
            error_action="ignore",
            suppress_warnings=True,
            stepwise=True,
        ),
        linear=dict(fit_intercept=True, normalize=False, n_jobs=None),
        svr=dict(kernel="rbf", C=1e3, gamma=0.1),
    )

    def __init_class_vars(
        self,
        prediction_type: str,
        dataset: pd.DataFrame,
        label: str,
        date_column: str,
        data_points: int,
    ):

        # either all to predict using all the models else one of arima, svr or linear
        self.prediction_type = prediction_type

        # input data
        if isinstance(dataset, pd.Series):
            self.input_df = dataset.to_frame()
        elif isinstance(dataset, pd.DataFrame):
            self.input_df = dataset
        else:
            raise ValueError("The dataset should either be a Dataframe or Series.")

        # the column name to be used as the date column
        self.date_column = date_column

        # The value to be predicted
        if label in self.input_df.columns:
            self.label = label
        else:
            raise ValueError("The label should a column name of the dataset.")

        # the number of data points in the future to be predicted, max 1000
        self.data_points = data_points

    def __createmodel(self, model_type, df):

        if model_type == "ARIMA":
            import pmdarima as pm

            model = pm.auto_arima(df[self.label], **self.model_params["arima"])
            predicted_values = model.predict(n_periods=self.data_points)
        elif model_type in ("LINEAR", "SVR"):
            prediction = (
                df[self.label].shift(-self.data_points).to_numpy().reshape(-1, 1)
            )
            X = df[self.label].to_numpy().reshape(-1, 1)[: -self.data_points]
            y = prediction[: -self.data_points].ravel()
            if model_type == "SVR":
                from sklearn.svm import SVR

                model = SVR(**self.model_params["svr"])
            else:
                from sklearn.linear_model import LinearRegression

                model = LinearRegression(**self.model_params["linear"])
            model.fit(X, y)
            x_predict = (
                df[self.label]
                .to_numpy()
                .reshape(-1, 1)[-self.data_points :]  # noqa: E203
            )
            predicted_values = model.predict(x_predict)
        else:
            raise ValueError(
                "Please specify an prediction_type as arima OR linear OR svr or all"
            )
        return model, [df[self.label].iloc[-1], *predicted_values]

    def __transform_output(self, data_df, predicted_values, predicted_col):
        predicted_date = pd.date_range(
            start=data_df[self.date_column].max(), periods=self.data_points + 1
        )
        predict_df = pd.DataFrame(data=predicted_values, columns=[predicted_col])
        predict_df[self.date_column] = predicted_date
        return predict_df

    def __modelling_prediction(self, df):
        if self.prediction_type == "all" or self.prediction_type == "COMPOUND":
            output_dfs = [
                df.copy(),
            ]
            models_dict = {}
            predicted_cols = []
            for param in self.model_params:
                model_type = self.param_model_map[param]

                predicted_col = model_type
                predicted_cols.append(predicted_col)
                model, predicted_values = self.__createmodel(model_type, df)
                models_dict[model_type] = model
                predicted_df = self.__transform_output(
                    df.copy(), predicted_values, predicted_col
                )
                output_dfs.append(predicted_df)
        else:
            output_dfs = [
                df.copy(),
            ]
            models_dict = {}
            predicted_cols = []
            model_type = self.param_model_map[self.prediction_type]
            predicted_col = model_type
            predicted_cols.append(predicted_col)
            model, predicted_values = self.__createmodel(model_type, df)
            models_dict[model_type] = model
            predicted_df = self.__transform_output(
                df.copy(), predicted_values, predicted_col
            )
            output_dfs.append(predicted_df)
        return predicted_cols, output_dfs

    def __multi_company(self):
        companies = self.input_df.Company.unique()
        output_dfs = []
        predicted_cols = []
        for company in companies:
            filtered = self.input_df.loc[self.input_df["Company"] == company]
            predicted_cols, outputs = self.__modelling_prediction(filtered)
            for out in outputs:
                out["Company"] = company
                out = out.reset_index(drop=True)
            output_dfs = [*output_dfs, *outputs]

        return predicted_cols, output_dfs

    def get(
        self,
        dataset: pd.DataFrame,
        prediction_type: str = "COMPOUND",
        label: str = "Close",
        date_column: str = "Date",
        data_points: int = 20,
    ):
        # initializes the class variables
        self.__init_class_vars(
            prediction_type=prediction_type,
            dataset=dataset,
            label=label,
            date_column=date_column,
            data_points=data_points,
        )

        # modelling and making the predictions
        output_dfs = None
        predicted_cols = None
        group = None
        if "Company" in dataset.columns:
            group = ["Date", "Company"]
            predicted_cols, output_dfs = self.__multi_company()
        else:
            group = ["Date"]
            predicted_cols, output_dfs = self.__modelling_prediction(dataset)
        res = pd.concat(output_dfs)
        res = res.reset_index(drop=True)
        agg = {i: ("sum" if i in predicted_cols else "first") for i in res.columns}
        res = res.groupby(group, as_index=False, dropna=False).agg(agg)
        if len(predicted_cols) > 1:
            res["COMPOUND"] = res[predicted_cols].mean(axis=1)
            res["COMPOUND"] = res["COMPOUND"].replace(0.000000, np.nan)
        for col in predicted_cols:
            res[col] = res[col].replace(0.000000, np.nan)
        if prediction_type == "COMPOUND":
            res = res.drop(columns=predicted_cols)
        return res



from naas_drivers.driver import InDriver
from newsapi.newsapi_client import NewsApiClient
import pandas as pd
import os


class Newsapi(InDriver):
    __key = None

    def connect(self, key=None):
        self.__key = key if key else os.environ.get("APINEW_KEY", None)
        self.connected = True
        return self

    def __transformDate(self, data, fields, limit):
        news = []
        for element in data:
            new_formated = {}
            for field in fields:
                if field == "title":
                    new_formated["title"] = element["title"]
                elif field == "link":
                    new_formated["link"] = element["url"]
                elif field == "description":
                    new_formated["description"] = element["description"]
                elif field == "source":
                    new_formated["source"] = element["source"]["name"]
                elif field == "image":
                    new_formated["image"] = element["urlToImage"]
                elif field == "author":
                    new_formated["author"] = element["author"]
                elif field == "date":
                    new_formated["date"] = element["publishedAt"]
                else:
                    raise ValueError("Unknow parameter")
            news.append(new_formated)

        if limit and isinstance(limit, int) and limit > 0:
            news = news[:limit]
        return pd.DataFrame.from_records(news)

    def get_sources(
        self, fields=["image", "title", "image", "link"], limit=20, **kargs
    ):
        self.check_connect()
        newsapi = NewsApiClient(api_key=self.__key)
        sources = newsapi.get_sources(**kargs)
        return self.__transformDate(sources.get("sources"), fields, limit)

    def get_top(self, fields=["image", "title", "image", "link"], limit=20, **kargs):
        self.check_connect()
        newsapi = NewsApiClient(api_key=self.__key)
        tops = newsapi.get_top_headlines(**kargs)
        return self.__transformDate(tops.get("articles"), fields, limit)

    def get(self, fields=["image", "title", "image", "link"], limit=20, **kargs):
        self.check_connect()
        newsapi = NewsApiClient(api_key=self.__key)
        all_news = newsapi.get_everything(**kargs)
        return self.__transformDate(all_news.get("articles"), fields, limit)



from naas_drivers.driver import InDriver, OutDriver
from airtable import Airtable as at
import pandas as pd


class Airtable(InDriver, OutDriver):
    _airtable = None
    _key = None
    _table = None

    def connect(self, key, table):
        self._key = key
        self._table = table
        self._airtable = at(key, table)
        self.connected = True
        return self

    def convert_data_to_df(self, data) -> pd.DataFrame:
        rows = []
        for dat in data:
            rows.append(
                {"id": dat["id"], "createdTime": dat["createdTime"], **dat["fields"]}
            )
        df = pd.DataFrame.from_records(rows)
        df = df.set_index("id")
        return df

    def get(self, **kwagrs):
        self.check_connect()
        data = self._airtable.get_all(**kwagrs)
        return self.convert_data_to_df(data)

    def send(self, **kwagrs):
        self.check_connect()
        return self._airtable.insert(**kwagrs)

    def search(self, **kwagrs):
        self.check_connect()
        data = self._airtable.search(**kwagrs)
        return self.convert_data_to_df(data)

    def update_by_field(self, **kwagrs):
        self.check_connect()
        return self._airtable.update_by_field(**kwagrs)

    def delete_by_field(self, **kwagrs):
        self.check_connect()
        return self._airtable.delete_by_field(**kwagrs)



import pandas as pd
from typing import List

# inspired from https://medium.com/bigdatarepublic/advanced-pandas-optimize-speed-and-memory-a654b53be6c2


class Optimise:
    def optimize_floats(self, df: pd.DataFrame) -> pd.DataFrame:
        floats = df.select_dtypes(include=["float64"]).columns.tolist()
        df[floats] = df[floats].apply(pd.to_numeric, downcast="float")
        return df

    def optimize_ints(self, df: pd.DataFrame) -> pd.DataFrame:
        ints = df.select_dtypes(include=["int64"]).columns.tolist()
        df[ints] = df[ints].apply(pd.to_numeric, downcast="integer")
        return df

    def optimize_objects(
        self, df: pd.DataFrame, datetime_features: List[str]
    ) -> pd.DataFrame:
        for col in df.select_dtypes(include=["object"]):
            if col not in datetime_features:
                num_unique_values = len(df[col].unique())
                num_total_values = len(df[col])
                if float(num_unique_values) / num_total_values < 0.5:
                    df[col] = df[col].astype("category")
            else:
                df[col] = pd.to_datetime(df[col])
        return df

    def get(self, df: pd.DataFrame, datetime_features: List[str] = []):
        return self.optimize_floats(
            self.optimize_ints(self.optimize_objects(df, datetime_features))
        )



import ssl
import ftplib
import datetime
import os


class Ftpsbase(ftplib.FTP_TLS):
    """FTPS subclass that automatically wraps sockets in SSL to support implicit FTPS."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._sock = None

    @property
    def sock(self):
        """Return the socket."""
        return self._sock

    @sock.setter
    def sock(self, value):
        """When modifying the socket, ensure that it is ssl wrapped."""
        if value is not None and not isinstance(value, ssl.SSLSocket):
            value = self.context.wrap_socket(value)
        self._sock = value

    def get_file(self, path, dest_path=None):
        """Read and save data from file which is specified 'path' """
        saved_path = None
        filename = os.path.basename(path)
        if filename != path:
            saved_path = self.pwd()
            self.cwd(os.path.dirname(path))
        with open(dest_path if dest_path else filename, "wb") as f:

            def callback(data):
                f.write(data)

            self.retrbinary(f"RETR {filename}", callback)
        if filename != path:
            self.cwd(saved_path)

    def send_file(self, path, dest_path=None):
        saved_path = None
        if dest_path is not None:
            saved_path = self.pwd()
            self.cwd(os.path.dirname(dest_path))
        filename = os.path.basename(path)
        ftp_command = f"STOR {filename}"
        file_byte = open(path, "rb")  # file to send
        self.storbinary(ftp_command, file_byte)  # send the file
        file_byte.close()
        if dest_path is not None:
            self.cwd(saved_path)

    def list_directory(self, dir_ame):
        dirs = []
        self.dir(dir_ame, dirs.append)
        files = []
        for dirr in dirs:
            line = dirr.lstrip().split()
            datetmp = " ".join(line[5 : len(line) - 1])  # noqa: E203
            month, day, year = datetmp.split(" ")
            time = None
            now = datetime.datetime.now()
            if year.find(":") > -1:
                time = year
                year = now.year
            else:
                time = "00:00"
            files.append(
                {
                    "permission": line[0],
                    "links": line[1],
                    "owner": line[2],
                    "group": line[3],
                    "size": line[4],
                    "month": month,
                    "day": day,
                    "year": year,
                    "time": time,
                    "name": line[-1],
                }
            )
        return files

    def help(self):
        print(f"=== {type(self).__name__} === \n")
        print(".get_file(path) => get file from ftp path\n")
        print(".send_file(path, dest_path) => send file to ftp path\n")
        print(".list_directory(path) => do ls in ftp in path\n")



from naas_drivers.driver import InDriver, OutDriver
from .__ftp import Ftpbase
from .__ftps import Ftpsbase


class Ftp(InDriver, OutDriver):
    """FTP subclass"""

    _ftp = None
    _user = None
    _passwd = None

    def connect(
        self,
        user,
        passwd,
        host="ftps.toucantoco.com",
        port=990,
        secure=False,
        force_prot=False,
    ):
        if secure:
            self._ftp = Ftpsbase()
        else:
            self._ftp = Ftpbase()
        self._ftp.connect(host=host, port=port)
        self._user = user
        self._passwd = passwd
        self._ftp.login(user=user, passwd=passwd)
        self.__force_prot = force_prot
        self.connected = True

    def __before_all(self):
        self.check_connect()
        if self.__force_prot:
            self._ftp.prot_c()
            self._ftp.prot_p()

    def get(self, path):
        self.__before_all()
        return self._ftp.get_file(path)

    def send(self, path, dest_path=None):
        self.__before_all()
        self._ftp.send_file(path, dest_path)

    def list(self, dir_name):
        self.__before_all()
        return self._ftp.list_directory(dir_name)



from naas_drivers.input.yahoofinance import Yahoofinance
from naas_drivers.input.cityfalcon import Cityfalcon
from naas_drivers.input.geolocator import Geolocator
from naas_drivers.input.newsapi import Newsapi
from naas_drivers.input.prediction import Prediction
from naas_drivers.input.sentiment import Sentiment
from naas_drivers.input.optimise import Optimise
from naas_drivers.input.pdf import Pdf
from naas_drivers.input.plotly import Plotly
from naas_drivers.input.html import Html
from naas_drivers.output.bubble import Bubble
from naas_drivers.output.email import Email
from naas_drivers.output.healthcheck import Healthcheck
from naas_drivers.output.ifttt import Ifttt
from naas_drivers.output.integromat import Integromat
from naas_drivers.output.zappier import Zappier
from naas_drivers.inputOutput.bobapp import Bobapp
from naas_drivers.inputOutput.airtable import Airtable
from naas_drivers.inputOutput.jupyter import Jupyter
from naas_drivers.inputOutput.ftp import Ftp
from naas_drivers.inputOutput.git import Git
from naas_drivers.inputOutput.gsheet import Gsheet
from naas_drivers.inputOutput.mongo import Mongo
from naas_drivers.inputOutput.toucan import Toucan
import requests
import os

__version__ = "0.25.4"

__github_repo = "jupyter-naas/drivers"

if os.environ.get("NAAS_DRIVER_LIGHT_INIT"):
    exit()

# In drivers
optimise = Optimise()
cityfalcon = Cityfalcon()
geolocator = Geolocator()
newsapi = Newsapi()
prediction = Prediction()
sentiment = Sentiment()
yahoofinance = Yahoofinance()
pdf = Pdf()
plotly = Plotly()
html = Html()

# Out drivers
bubble = Bubble()
email = Email()
healthcheck = Healthcheck()
ifttt = Ifttt()
integromat = Integromat()
zappier = Zappier()

# InOut drivers
bobapp = Bobapp()
airtable = Airtable()
jupyter = Jupyter()
ftp = Ftp()
git = Git()
gsheet = Gsheet()
mongo = Mongo()
toucan = Toucan()

__doc_url = "https://naas.gitbook.io/drivers/"


def doc():
    return __doc_url


def version():
    print(__version__)


def get_last_version():
    url = f"https://api.github.com/repos/{__github_repo}/tags"
    response = requests.get(url, headers={"Accept": "application/vnd.github.v3+json"})
    return response.json()[0]["name"]


def up_to_date():
    return get_last_version() == version()
